classdef TriElasticity < TriMesher
    properties(GetAccess = 'public', SetAccess = 'public')
        myBCtype;
        myBCvalue;
        myDOFPerNode;
        myNumDOF;
        myDOFPerElem;
        myClass;
        myE;
        myNu;
        myD;
        myBodyForce;
        myXi;
        myEta;
        myWt;
        myGradN;
        myK;
        myF;
        myC;
        myfC;
        mySol;
        myU;
        myV;
        myStrainElems;
        myStressElems;
        myPrincipalStressesElems;
        myPrincipalStressNodes;
        myStrainEnergyElems;
        myVonMisesElems;
        myVonMisesNodes;
        myStrainEnergyDensityNodes;
        myCompliance;
        myElementArea;
        myFixedDOF; % fixed degrees of freedom
        myFreeDOF; % all free dof
        myForcedNodes; % with non-zero forces
        myFixedNodes;
        mySolverMethod;  % solver methods
        myMaxDelta; % maximum displacement
        myMaxStress; % max von Mises stress in the domain at any instance
        myN;
    end
    methods
        % default: use the parent constructor
        function obj = TriElasticity(brepFileName,nElements,class) 
            if (nargin < 3)
                class = 'PlaneStress';
            end 
            if (nargin < 2)
                class = 'PlaneStress';
                nElements = 500;
            end 
            obj = obj@TriMesher(brepFileName,nElements); % call superclass 
            % set the default boundary Conditions
            obj.myBCtype = zeros(obj.myNumBoundarySegments,2);% u v
            obj.myBCvalue = zeros(obj.myNumBoundarySegments,2);% u v
            obj.myDOFPerNode = 2; % (u, v)
            obj.myDOFPerElem = obj.myDOFPerNode*obj.myNodesPerElement;
            obj.myNumDOF = obj.myDOFPerNode*obj.myNumNodes;   
            obj.myE = 2e11;
            obj.myNu = 0.33;  
            E = obj.myE;
            nu = obj.myNu;
            obj.myBodyForce = [0 0];
            obj.myClass = class;

            obj.mySolverMethod = 2; 
            if (strcmp(obj.myClass,'PlaneStrain'))
                obj.myD = E/((1+nu)*(1-2*nu))*[1-nu nu 0; nu 1-nu 0;0 0 (1-2*nu)/2];
            elseif (strcmp(obj.myClass,'PlaneStress'))
                obj.myD = E/(1-nu^2)*[1 nu 0; nu 1 0;0 0 (1-nu)/2];
            elseif (strcmp(obj.myClass,'AxiSymmetric'))
                obj.myD = E/((1+nu)*(1-2*nu))*[1-nu nu nu 0; nu 1-nu nu 0; nu nu 1-nu 0;0 0 0 (1-2*nu)/2];
            else
                disp('Class undefined; options: PlaneStrain, PlaneStress, AxisSymmetric ');
                return;
            end
        end
        function obj = readMesh(obj,fileName)
            obj = obj.readMesh@TriMesher(fileName);
            obj.myNumDOF = obj.myDOFPerNode*obj.myNumNodes;   
        end
        function obj = readOffFileMesh(obj,fileName)
            obj = obj.readOffFileMesh@TriMesher(fileName);
            obj.myNumDOF = obj.myDOFPerNode*obj.myNumNodes;   
        end
        function obj = resetBrepAndSolve(obj,brep)
            % Useful during shape optimization
            obj = obj.resetBrepAndMesh(brep);
            obj.myNumDOF = obj.myDOFPerNode*obj.myNumNodes;   
            obj = obj.solveLinearElasticityProblem(); % Solve Primary FEA problem
        end
        function obj = setSolverMethod(obj,method)
            obj.mySolverMethod = method;
        end  
        function obj = setYoungsModulus(obj,E)
            obj.myE = E;
            obj = obj.computeDMatrix();
        end
        function obj = setPoissonsRatio(obj,nu)
            obj.myNu = nu;
            obj = obj.computeDMatrix();
        end
        
        function obj = computeDMatrix(obj)
            E = obj.myE;
            nu = obj.myNu;
            if (strcmp(obj.myClass,'PlaneStrain'))
                obj.myD = E/((1+nu)*(1-2*nu))*[1-nu nu 0; nu 1-nu 0;0 0 (1-2*nu)/2];
            elseif (strcmp(obj.myClass,'PlaneStress'))
                obj.myD = E/(1-nu^2)*[1 nu 0; nu 1 0;0 0 (1-nu)/2];
            elseif (strcmp(obj.myClass,'AxiSymmetric'))
                obj.myD = E/((1+nu)*(1-2*nu))*[1-nu nu nu 0; nu 1-nu nu 0; nu nu 1-nu 0;0 0 0 (1-2*nu)/2];
            else
                disp('Class undefined ');
                return;
            end
        end
        
        function obj = applyXForceOnEdge(obj,boundaryEdges,force)
            % convert force into pressure
            obj.myBCvalue(boundaryEdges,1) = force./obj.mySegmentLengths(boundaryEdges);
        end
        function obj = applyYForceOnEdge(obj,boundaryEdges,force)
            % convert force into pressure
            obj.myBCvalue(boundaryEdges,2) = force./obj.mySegmentLengths(boundaryEdges);
        end
        function obj = fixXOfEdge(obj,boundaryEdges)
            obj.myBCtype(boundaryEdges,1) = 1;
        end
        function obj = fixYOfEdge(obj,boundaryEdges)
            obj.myBCtype(boundaryEdges,2) = 1;
        end
        function obj = fixEdge(obj,boundaryEdges)
            obj.myBCtype(boundaryEdges,1:2) = 1;
        end
        function obj = setForceVector(obj,f)
            obj.myF = f;
        end
        function obj = applyBodyForce(obj,b)
           obj.myBodyForce = b;
        end
        function [udof,vdof] = getDof(obj,elem)
            nodes = obj.myMesh.t(1:obj.myNodesPerElement,elem)';
            udof = 2*nodes-1;
            vdof = 2*nodes;
        end
        
        function obj  = assembleK(obj)
            % Assemble K and f (without boundary conditions) for 2D Linear Elasticity
            % mapping is needed when a subset of elements need to be assembled.
            nDOF = obj.myNumNodes*2;
            if (strcmp(obj.myShape,'Linear'))
                [xi_GQ,eta_GQ,wt_GQ]= obj.GaussQTriangle(1);
            elseif (strcmp(obj.myShape,'Quadratic'))
                [xi_GQ,eta_GQ, wt_GQ]= obj.GaussQTriangle(3);
            end
            obj.myXi = xi_GQ;
            obj.myEta = eta_GQ;
            obj.myWt = wt_GQ;
           
            obj.myDOFPerElem = obj.myDOFPerNode*obj.myNodesPerElement;
            nTriangles = obj.myNumElems;
            nzmax = obj.myDOFPerElem^2*nTriangles;
            RowTriplets = zeros(nzmax,1);
            ColTriplets = zeros(nzmax,1);
            EntryTriplets = zeros(nzmax,1);
            f = zeros(nDOF,1);
            obj.myN = cell(1,length(xi_GQ));
            obj.myGradN = cell(1,length(xi_GQ));
            for i = 1:length(xi_GQ)
                [obj.myN{i},obj.myGradN{i}] = obj.triShapeFunction(xi_GQ(i),eta_GQ(i));
            end
            index = 1;
            for elem = 1:nTriangles
                nodes = obj.myMesh.t(1:obj.myNodesPerElement,elem)';
                [KElem,fElem,AElem] = obj.integrateKOverElem(elem);     
                if (obj.myPseudoDensity(elem) == 0) % during topology optimization
                    KElem = eye(obj.myDOFPerElem); %avoids singularity in matrix
                    fElem = zeros(obj.myDOFPerElem,1);
                else
                    KElem = obj.myPseudoDensity(elem)^3*KElem;
                    fElem = obj.myPseudoDensity(elem)*fElem;
                end
                obj.myElementArea(elem)= AElem;
                dof = [2*nodes-1; 2*nodes];
                dof = reshape(dof,1,obj.myDOFPerElem);
                temp = dof(ones(1,obj.myDOFPerElem),:);
                colIndex = reshape(temp',1,obj.myDOFPerElem^2);
                rowIndex = reshape(temp,1,obj.myDOFPerElem^2);
                entries = reshape(KElem',1,obj.myDOFPerElem^2);
                RowTriplets(index:index+obj.myDOFPerElem^2-1,1) = rowIndex';
                ColTriplets(index:index+obj.myDOFPerElem^2-1,1) = colIndex';
                EntryTriplets(index:index+obj.myDOFPerElem^2-1,1) = entries';
                index = index+obj.myDOFPerElem^2;
                f(dof) = f(dof) + fElem;
            end
            obj.myK = sparse(RowTriplets,ColTriplets,EntryTriplets,nDOF,nDOF);
            obj.myF = f;
        end
        function [KElem,fElem] = computeElementStiffness(obj,elem)
            [KElem,fElem] = integrateKOverElem(obj,elem);
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function [KElem,fElem,AElem] = integrateKOverElem(obj,elem)
            
            NCell = obj.myN;
            gradNCell = obj.myGradN;
            nodes = obj.myMesh.t(1:obj.myNodesPerElement,elem)';
            xNodes = obj.myMesh.p(1,nodes);
            yNodes = obj.myMesh.p(2,nodes);
            KElem = zeros(obj.myDOFPerElem,obj.myDOFPerElem);
            fElem = zeros(obj.myDOFPerElem,1);
            % Since we are using linear interpolation for the geometry
            % the Jacobian is a constant.
            invJ = [(-yNodes(1)+yNodes(3)) (-yNodes(2)+yNodes(1)); ...
                (-xNodes(3)+xNodes(1)) (-xNodes(1)+xNodes(2))];
            dJ = invJ(1,1)*invJ(2,2)-invJ(1,2)*invJ(2,1);
            AElem = dJ/2;
            invJ = invJ/dJ;
            Z = zeros(1,obj.myDOFPerElem/2);
            wt_GQ = obj.myWt;
            bx = obj.myBodyForce(1);
            by = obj.myBodyForce(2);
            for g = 1:length(wt_GQ)
                N = NCell{g};
                gradN = gradNCell{g};
                x = xNodes*N;% note: for axisymmetric, x is the radius
               
                T1 = invJ(1,:)*gradN;
                T2 = invJ(2,:)*gradN;
                D = obj.myD;
                if (strcmp(obj.myClass,'PlaneStrain'))
                    B = [T1 Z; Z T2; T2 T1];
                    KElem = KElem + wt_GQ(g)*dJ*B'*D*B;
                    fElem = fElem + wt_GQ(g)*dJ*[N*bx;N*by];
                elseif (strcmp(obj.myClass,'PlaneStress'))
                    B = [T1 Z; Z T2; T2 T1];
                    KElem = KElem + wt_GQ(g)*dJ*B'*D*B;
                    fElem = fElem + wt_GQ(g)*dJ*[N*bx;N*by];
                elseif (strcmp(obj.myClass,'AxiSymmetric'))
                    B = [T1 Z; Z T2;N'/x Z; T2 T1]; % note: for axisymmetric, x is the radius
                    KElem = KElem + x*wt_GQ(g)*dJ*B'*D*B;
                    fElem = fElem + x*wt_GQ(g)*dJ*[N*bx;N*by];
                end
            end
            if (obj.myDOFPerElem == 6)
                order = reshape([1:3;4:6],1,6);%order = [1 4 2 5 3 6];
            elseif (obj.myDOFPerElem == 12)
                order = reshape([1:6;7:12],1,12);
            end
            KElem = KElem(order,order);
            fElem = fElem(order);
        end
         %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function obj = assembleBC(obj)
            if (strcmp(obj.myShape,'Linear'))
                [xi_GQ, wt_GQ] = obj.GaussQLine(1);
                N1D = cell(1,length(xi_GQ));
                for i = 1:length(xi_GQ)
                    N1D{i} = obj.edgeShapeFunction(xi_GQ(i));
                end
            elseif (strcmp(obj.myShape,'Quadratic'))
                [xi_GQ, wt_GQ] = obj.GaussQLine(3);
                N1D = cell(1,length(xi_GQ));
                for i = 1:length(xi_GQ)
                    temp = obj.edgeShapeFunction(xi_GQ(i));
                    N1D{i} = temp([1 3 2]); % need to change the order
                end
            end 
            nDOF = obj.myNumDOF;
            % Assemble surface force (Neumann data)
            fBoundary = zeros(nDOF,1);
            obj.myForcedNodes = [];
            for geomEdge = 1:size(obj.myBrep.segments,2)
                typeu = obj.myBCtype(geomEdge,1);
                typev = obj.myBCtype(geomEdge,2);
                valueu = obj.myBCvalue(geomEdge,1);
                valuev = obj.myBCvalue(geomEdge,2);
                if (typeu == 0 ) && (abs(valueu) > 0)
                    boundarySegments = find(obj.myMesh.e(5,:) == geomEdge);
                    for seg = boundarySegments
                        nodes = obj.myMesh.e(1:obj.myNodesPerEdge,seg);
                        udof = 2*nodes-1;
                        fBoundaryElem = obj.integrateOverBoundary(geomEdge,seg,wt_GQ,N1D,1);
                        fBoundary(udof) = fBoundary(udof) +fBoundaryElem;
                        obj.myForcedNodes = unique([obj.myForcedNodes ;nodes]);
                    end
                end
                if (typev == 0 ) && (abs(valuev) > 0)
                    boundarySegments = find(obj.myMesh.e(5,:) == geomEdge);
                    for seg = boundarySegments
                        nodes = obj.myMesh.e(1:obj.myNodesPerEdge,seg);
                        vdof = 2*nodes;
                        fBoundaryElem = obj.integrateOverBoundary(geomEdge,seg,wt_GQ,N1D,2);
                        fBoundary(vdof) = fBoundary(vdof) + fBoundaryElem;
                        obj.myForcedNodes = unique([obj.myForcedNodes ;nodes]);
                    end
                end
            end
            % Gather Dirichlet boundary conditions
            isDirichlet = zeros(nDOF,1);
            dirValue = zeros(nDOF,1);
            for geomEdge = 1:size(obj.myBrep.segments,2)
                typeu = obj.myBCtype(geomEdge,1);
                typev = obj.myBCtype(geomEdge,2);
                valueu = obj.myBCvalue(geomEdge,1);
                valuev = obj.myBCvalue(geomEdge,2);
                if (typeu == 1)
                    boundarySegments = find(obj.myMesh.e(5,:) == geomEdge);
                    for seg = boundarySegments
                        nodes = obj.myMesh.e(1:obj.myNodesPerEdge,seg);
                        udof = 2*nodes-1;
                        isDirichlet(udof) = 1;
                        dirValue(udof) =  valueu;
                    end
                end
                if (typev == 1)
                    boundarySegments = find(obj.myMesh.e(5,:) == geomEdge);
                    for seg = boundarySegments
                        nodes = obj.myMesh.e(1:obj.myNodesPerEdge,seg);
                        vdof = 2*nodes;
                        isDirichlet(vdof) = 1;
                        dirValue(vdof) =  valuev;
                    end
                end
            end
            dirichletDOF = find(isDirichlet(:) == 1);
            obj.myFixedDOF = dirichletDOF;
            nDirichletDOF = length(dirichletDOF);
            C = zeros(nDirichletDOF,nDOF);
            for i = 1:nDirichletDOF
                C(i,dirichletDOF(i)) = 1;
            end
            C = sparse(C);
            fDirichlet = dirValue(dirichletDOF);
            obj.myF = obj.myF + fBoundary;
            obj.myC = C;
            obj.myfC = fDirichlet;
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function fBoundaryElem = integrateOverBoundary(obj,geomEdge,seg,wt_GQ,N1D,dof)
            nodes = obj.myMesh.e(1:obj.myNodesPerEdge,seg);
            xNodes = obj.myMesh.p(1,nodes);
            yNodes = obj.myMesh.p(2,nodes);
            dx = xNodes(2)-xNodes(1);
            dy = yNodes(2)-yNodes(1);
            L = sqrt(dx^2 + dy^2);
            vec = [dx dy 0]/L;
            zVec = [0 0 1];
            normal = cross(vec,zVec);
            nx = normal(1); %#ok<NASGU>
            ny = normal(2); %#ok<NASGU>
            fBoundaryElem = zeros(numel(N1D{1}),1);
            for g = 1:length(wt_GQ)
                N = N1D{g};
                x = xNodes*N; % radius for axisymmetric problems
                f =  obj.myBCvalue(geomEdge,dof); % dof is either 1 (u) or 2 (v)
                if (strcmp(obj.myClass,'PlaneStrain')) || (strcmp(obj.myClass,'PlaneStress'))
                    fBoundaryElem = fBoundaryElem + wt_GQ(g)*(L/2)*N*f;
                elseif (strcmp(obj.myClass,'AxiSymmetric'))
                    fBoundaryElem = fBoundaryElem + x*wt_GQ(g)*(L/2)*N*f;
                end
            end
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function [obj] = assemble(obj)
            obj = obj.assembleK();
            obj = obj.assembleBC();
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function [obj,success] = solve(obj)
            [obj,success] = obj.solveLinearSystem();
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function [obj] = setConstraintMatrix(obj,C)
            obj.myC = C;
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function [obj] = setDirichletCondition(obj,dirichletConditions)
            obj.myfC = dirichletConditions(:);
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function [obj,success] = solveLinearSystem(obj)
            allDOF = 1:obj.myNumDOF;
            method = obj.mySolverMethod;
            success = 2;
            if (method == 1) % direct solve with Lagrange multipliers
                scale = max(max(obj.myK));
                obj.myC = scale*obj.myC;
                obj.myfC = scale*obj.myfC;
                [nDirichletDOF] = size(obj.myC,1);
                % Solution of Algebraic Problem
                Z = spalloc(nDirichletDOF,nDirichletDOF,1);
                KBar = sparse([obj.myK obj.myC'; obj.myC Z]);
                fBar = sparse([obj.myF;  obj.myfC]);
                soln = KBar \ fBar;
            elseif (method == 2)% direct solve with elimination  of fixed dof
                obj.myFreeDOF  = setdiff(allDOF,obj.myFixedDOF);
                % the useful part of the KMatrix
                KTilde = obj.myK(obj.myFreeDOF,obj.myFreeDOF);
                fTilde = obj.myF(obj.myFreeDOF);
                % now subtract all the dirichlet values from rhs
                for i = 1:numel(obj.myFixedDOF)
                    dof = obj.myFixedDOF(i);
                    if (abs(obj.myfC(i)) > 0)
                        fTilde = fTilde - obj.myK(obj.myFreeDOF,dof)*obj.myfC(i);
                    end
                end
                soln = zeros(obj.myNumDOF,1);
                soln(obj.myFixedDOF) = obj.myfC; % fixed values
                soln(obj.myFreeDOF) =  KTilde\fTilde;
            elseif (method == 3)% iterative solve with elimination of fixed dof
                obj.myFreeDOF  = setdiff(allDOF,obj.myFixedDOF);
                % the useful part of the KMatrix
                KTilde = obj.myK(obj.myFreeDOF,obj.myFreeDOF);
                fTilde = obj.myF(obj.myFreeDOF);
                % now subtract all the dirichlet values from rhs
                for i = 1:numel(obj.myFixedDOF)
                    dof = obj.myFixedDOF(i);
                    if (abs(obj.myfC(i)) > 0)
                        fTilde = fTilde - obj.myK(obj.myFreeDOF,dof)*obj.myfC(i);
                    end
                end
                soln = zeros(obj.myNumDOF,1);
                soln(obj.myFixedDOF) = obj.myfC; % fixed values
                [soln(obj.myFreeDOF),~, flag] =  obj.CG(KTilde,fTilde);
                if (flag == -1)
                    success = -1;
                    return;
                end
            end
            nDOF = size(obj.myK,1);
            obj.mySol = full(soln(1:nDOF));
            obj.myU = full(soln(1:2:nDOF));
            obj.myV = full(soln(2:2:nDOF));
            obj.myMaxDelta = max(sqrt(obj.myU.^2 + obj.myV.^2));
            obj.myCompliance = (obj.mySol)'*obj.myF;
            obj = obj.computeStresses();
            obj.myMaxStress = max(obj.myVonMisesElems);
        end
        function J = computeCompliance(obj)
             J = (obj.mySol)'*obj.myF;
        end
        function [sol,nodes,dof] = getElementSolution(obj,elem)
            nodes = obj.myMesh.t(1:obj.myNodesPerElement,elem);
            dof = [2*nodes-1; 2*nodes];
            n = obj.myNodesPerElement;
            order = reshape([1:n;n+1:2*n],1,2*n);
            dof = dof(order);% alternate u,v
            sol = obj.mySol(dof);
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function obj = computeStresses(obj)
            % Compute stresses at the average of all Gauss points
            nTriangles = obj.myNumElems;
            nNodes = obj.myNumNodes;
            %[~,gradN] = obj.triShapeFunction(1/3,1/3); % gradient at center
            %gradN = obj.myGradN;
            obj.myStrainElems = zeros(nTriangles,2,2);
            obj.myStressElems = zeros(nTriangles,2,2);
            obj.myStrainEnergyElems = zeros(nTriangles,1);
            obj.myVonMisesElems = zeros(1,nTriangles);
            obj.myVonMisesNodes = zeros(1,nNodes);
            obj.myStrainEnergyDensityNodes = zeros(1,nNodes);
            nElemsConnectedToNode  = zeros(1,nNodes);
            areasOfElemsConnectedToNode = zeros(1,nNodes); 
            D = obj.myD;
            gradNCell = obj.myGradN;
            for elem = 1:nTriangles
                if (obj.myPseudoDensity(elem) == 0)
                    continue;
                end
                obj.myVonMisesElems(elem) = 0;
                nodes = obj.myMesh.t(1:obj.myNodesPerElement,elem)';
                nElemsConnectedToNode(nodes) = nElemsConnectedToNode(nodes) + 1;
                areasOfElemsConnectedToNode(nodes) = areasOfElemsConnectedToNode(nodes) + obj.myElementArea(elem);
                xNodes = obj.myMesh.p(1,nodes);
                yNodes = obj.myMesh.p(2,nodes);
                uvalue = obj.myU(nodes);
                vvalue = obj.myV(nodes);
                invJ = [(-yNodes(1)+yNodes(3)) (-yNodes(2)+yNodes(1)); ...
                    (-xNodes(3)+xNodes(1)) (-xNodes(1)+xNodes(2))];
                dJ = invJ(1,1)*invJ(2,2)-invJ(1,2)*invJ(2,1);
                invJ = invJ/dJ;
                for i = 1:numel(gradNCell)
                    B = invJ*gradNCell{i};
                    gradu = B*uvalue;
                    gradv = B*vvalue;
                    ux = gradu(1);
                    uy = gradu(2);
                    vx = gradv(1);
                    vy = gradv(2);
                    strainElems = [ux (uy+vx)/2; (uy+vx)/2 vy];
                    sxx = D(1,1)*strainElems(1,1) + D(1,2)*strainElems(2,2);
                    syy = D(2,1)*strainElems(1,1) + D(2,2)*strainElems(2,2);
                    sxy = 2*D(3,3)*strainElems(1,2);
                    maxVSStress = sqrt(sxx*sxx + syy*syy - sxx*syy + 3*sxy*sxy);
                    if (maxVSStress > obj.myVonMisesElems(elem))
                        obj.myStrainElems(elem,:,:) = strainElems;
                        obj.myStressElems(elem,:,:) = [sxx sxy; sxy syy];
                        obj.myVonMisesElems(elem) =  maxVSStress;
                    end
                end
                % [ 1 2 4] to avoid double counting shear
                obj.myStrainEnergyElems(elem,1) = obj.myStrainElems(elem,[1 2 4])*obj.myStressElems(elem,[1 2 4])';
                obj.myVonMisesNodes(nodes) = obj.myVonMisesNodes(nodes) + obj.myVonMisesElems(elem);
                obj.myStrainEnergyDensityNodes(nodes) = obj.myStrainEnergyDensityNodes(nodes) + obj.myStrainEnergyElems(elem);
            end
            obj.myVonMisesNodes =  obj.myVonMisesNodes./nElemsConnectedToNode;
            obj.myStrainEnergyDensityNodes =  obj.myStrainEnergyDensityNodes./areasOfElemsConnectedToNode;
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function maxVonMisesStress = getMaxVonMisesStress(obj)
            maxVonMisesStress = max(obj.myVonMisesElems);
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function pNormStress = getPNormStress(obj,pNorm)
            if (nargin == 1)
                pNorm = 4;
            end
            pNormStress = (sum(obj.myVonMisesElems.^pNorm)).^(1/pNorm);
        end
         %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function obj = computePrincipalStresses(obj)
            % Compute principal stresses at each element
            nTriangles = obj.myNumElems;
            nNodes = obj.myNumNodes;
            obj.myPrincipalStressesElems = zeros(nTriangles,2);
            obj.myPrincipalStressNodes = zeros(nNodes,2);
            nElemsConnectedToNode = zeros(nNodes,1);
            for elem = 1:nTriangles
                if (obj.myPseudoDensity(elem) == 0)
                    continue;
                end
                nodes = obj.myMesh.t(1:obj.myNodesPerElement,elem)';
                nElemsConnectedToNode(nodes) = nElemsConnectedToNode(nodes) + 1;
                stress = obj.myStressElems(elem,1:2,1:2);
                stress = reshape(stress,[2 2]);
                obj.myPrincipalStressesElems(elem,:) = eig(stress)';
                obj.myPrincipalStressNodes(nodes,1) = obj.myPrincipalStressNodes(nodes,1) + obj.myPrincipalStressesElems(elem,1);
                obj.myPrincipalStressNodes(nodes,2) = obj.myPrincipalStressNodes(nodes,2) + obj.myPrincipalStressesElems(elem,2);
            end
            obj.myPrincipalStressNodes(:,1) =  obj.myPrincipalStressNodes(:,1)./nElemsConnectedToNode;
            obj.myPrincipalStressNodes(:,2) =  obj.myPrincipalStressNodes(:,2)./nElemsConnectedToNode;
        end
        function plotMesh(obj)
            obj.plotMesh@TriMesher();
            hold on;
            % mark all fixed nodes
            index = (rem(obj.myFixedDOF,2)==1);
            fixedXNodes = (obj.myFixedDOF(index)-1)/2+1;
            index = (rem(obj.myFixedDOF,2)==0);
            fixedYNodes = (obj.myFixedDOF(index))/2;
            onlyXFixed = setdiff(fixedXNodes,fixedYNodes);
            onlyYFixed = setdiff(fixedYNodes,fixedXNodes);
            bothFixed = intersect(fixedXNodes,fixedYNodes);
            plot(obj.myMesh.p(1,onlyXFixed),obj.myMesh.p(2,onlyXFixed),'>k','MarkerFaceColor','k');
            plot(obj.myMesh.p(1,onlyYFixed),obj.myMesh.p(2,onlyYFixed),'^k','MarkerFaceColor','k'); 
            plot(obj.myMesh.p(1,bothFixed),obj.myMesh.p(2,bothFixed),'ok','MarkerFaceColor','k');  
            
            % mark all forced nodes
            nodes = obj.myForcedNodes;
            Fx = obj.myF(2*nodes-1);
            Fy = obj.myF(2*nodes);
            xRange = max(obj.myMesh.p(1,:))-min(obj.myMesh.p(1,:));
            yRange = max(obj.myMesh.p(2,:))-min(obj.myMesh.p(2,:));
            scale = 0.25*min(xRange,yRange);
            normF = sqrt(Fx.^2 + Fy.^2);
            if (normF > 0)
                Fx = Fx./normF;
                Fy = Fy./normF;
                Fx = Fx(:)';
                Fy = Fy(:)';
                %plot(obj.myMesh.p(1,nodes),obj.myMesh.p(2,nodes),'dr');
                start = [obj.myMesh.p(1,nodes); obj.myMesh.p(2,nodes)];
                stop = start + scale*[Fx; Fy];
                obj.drawArrow(start',stop');
            end
            axis auto;axis equal;
            obj.adjustFigScale();
            str = [obj.myTitle ' #Elems = ' num2str(obj.myNumElems)];
            title(str); 
        end
        function plotDeformation(obj)
            scale = 0.1*obj.myModelScale/obj.myMaxDelta;
            elemsToPlot = obj.myPseudoDensity > 0;
            mesh = obj.myMesh;
            p = mesh.p + scale*[obj.myU';obj.myV'];
            t = mesh.t(1:3,elemsToPlot);
            t(4,:) = 0;
            pdemesh(p,mesh.e,t);   
            title(['Max Delta = ' num2str(obj.myMaxDelta)]);
            obj.adjustFigScale();
            axis equal;
        end     
        function plotStress(obj,scale)
            x = obj.myMesh.p(1,:);
            y = obj.myMesh.p(2,:);
            u = obj.myU;
            v = obj.myV;
            delta = max(sqrt(u.^2 + v.^2));
            modelScale = obj.myModelScale;
            if (nargin <= 2) % autoscale
                scale = 0.1*modelScale/(delta+eps);
            end
            x  = x + scale*u';
            y = y + scale*v';
            vonMises = obj.myVonMisesNodes;
            elemsToPlot = obj.myPseudoDensity > 0;
            trisurf(obj.myMesh.t(1:3,elemsToPlot)',x,y,vonMises,vonMises,'facecolor','interp');
            %view(2);axis equal;
            
            caxis([0 max(vonMises)]); colormap('default');
            colorbar('location','eastoutside')
            title(['\delta = ' num2str(obj.myMaxDelta,'%.2e') ' \sigma: ' num2str(max(vonMises),'%.2e')],'fontsize',18);
            axis equal;axis tight; grid off;
            obj.adjustFigScale();grid off;
            view(2);
        end
        function plotFirstPrincipalStress(obj)
            x = obj.myMesh.p(1,:);
            y = obj.myMesh.p(2,:);
            data = obj.myPrincipalStressNodes(:,1);
            elemsToPlot = obj.myPseudoDensity > 0;
            trisurf(obj.myMesh.t(1:3,elemsToPlot)',x,y,data,data,'facecolor','interp','edgecolor','none');
            view(2);axis('equal');
            obj.adjustFigScale();grid off;
            caxis([min(data) max(data)]); colormap('default');
            colorbar('location','eastoutside')
            title('FirstPrincipalStress');
        end
        function plotSecondPrincipalStress(obj)
            x = obj.myMesh.p(1,:);
            y = obj.myMesh.p(2,:);
            data = obj.myPrincipalStressNodes(:,2);
            elemsToPlot = obj.myPseudoDensity > 0;
            trisurf(obj.myMesh.t(1:3,elemsToPlot)',x,y,data,data,'facecolor','interp','edgecolor','none');
            view(2);axis('equal');
            obj.adjustFigScale();grid off;
            caxis([min(data) max(data)]); colormap('default');
            colorbar('location','eastoutside')
            title('SecondPrincipalStress');
        end
        function plotPrincipalStressContours(obj)
            sigma1 = obj.myPrincipalStressNodes(:,1);
            sigma2 = obj.myPrincipalStressNodes(:,2);
            elemsToPlot = obj.myPseudoDensity > 0;
            tricontour(obj.myMesh.p',obj.myMesh.t(1:3,elemsToPlot)',sigma1,20);      
            view(2);axis('equal');hold on;
            tricontour(obj.myMesh.p',obj.myMesh.t(1:3,elemsToPlot)',sigma2,20);
            obj.adjustFigScale();grid off;
            title('Principal Stresses Contours');
        end
        function plotStrainEnergyDensity(obj)
            x = obj.myMesh.p(1,:);
            y = obj.myMesh.p(2,:);
            energy = obj.myStrainEnergyDensityNodes;
            elemsToPlot = obj.myPseudoDensity > 0;
            trisurf(obj.myMesh.t(1:3,elemsToPlot)',x,y,energy,energy,'facecolor','interp','edgecolor','none');
            view(2);axis('equal');
            obj.adjustFigScale();grid off;
            caxis([0 max(energy)]); colormap('default');
            colorbar('location','eastoutside')
            title(['StrainEnergyDensity: ' num2str(max(energy))]);
        end
    end
end
